// Game:           Prince of Persia (SNES)
// Category:       100%
// Emulator:       snes9x-1.60-win32-x64
// Original by:    David from forum.princed.org
// Updated:        2023-02-26
// IGT timing:     YES

// Base memory path:
// snes9x-1.60-win32-x64: "snes9x-x64.exe"+008D8C38

// Known bugs:
// * The timer will not start if the player fast-forwards through START GAME.

// Open questions:
// * What if the player (dies and) restarts a level?
// * How to measure the time on level 20?
//     * The main timer stops there.
//     * The per-level time is reset after you defeat Jaffar and return to the hall. This is reflected on the BEST TIME screen.

state("snes9x-x64")
{
	byte720 LevelObjects          : "snes9x-x64.exe", 0x008D8C38, 0x1E170; // 7F:E170; 24 rooms * 3 rows * 10 columns
	byte NextLevel                : "snes9x-x64.exe", 0x008D8C38, 0x0544; // 7E:0544; 0 means LEVEL 1, 1 means LEVEL 2, and so on.
	byte CurrentLevel             : "snes9x-x64.exe", 0x008D8C38, 0x0579; // 7E:0579; 0 means LEVEL 1, 1 means LEVEL 2, and so on.
	//byte CharScrn                 : "snes9x-x64.exe", 0x008D8C38, 0x0462; // 7E:0462
	byte KidMaxHP                 : "snes9x-x64.exe", 0x008D8C38, 0x050A; // 7E:050A
	ushort FrameCount             : "snes9x-x64.exe", 0x008D8C38, 0x052D; // 7E:052D; elapsed frames (8 video frames each); stops on level 20
	byte FrameCountHalf           : "snes9x-x64.exe", 0x008D8C38, 0x052F; // 7E:052F; elapsed half frames (0 or 1) (4 video frames each)
	ushort FrameCountThisLevel    : "snes9x-x64.exe", 0x008D8C38, 0x054B; // 7E:054B; elapsed frames on this level
	byte FrameCountHalfThisLevel  : "snes9x-x64.exe", 0x008D8C38, 0x054D; // 7E:054D; elapsed half frames (0 or 1) + overflow bits multiplied by 2
	ushort FrameCountLevelStart   : "snes9x-x64.exe", 0x008D8C38, 0x054E; // 7E:054E; elapsed frames when the level started; does not work on level 1!
	byte FrameCountHalfLevelStart : "snes9x-x64.exe", 0x008D8C38, 0x0550; // 7E:0550; elapsed half frames (0 or 1) + overflow bits multiplied by 2
	byte StartRoom                : "snes9x-x64.exe", 0x008D8C38, 0x1F110; // 7F:F110
	byte Kid_room                 : "snes9x-x64.exe", 0x008D8C38, 0x0472; // 7E:0472

	sbyte Guard_alive             : "snes9x-x64.exe", 0x008D8C38, 0x0486; // 7E:0486; < 0 if alive, >= 0 if dead
	byte Guard_room               : "snes9x-x64.exe", 0x008D8C38, 0x0482; // 7E:0482
	byte Guard_direction          : "snes9x-x64.exe", 0x008D8C38, 0x047A; // 7E:047A; 0x7F if there is no guard
	byte Guard_hp                 : "snes9x-x64.exe", 0x008D8C38, 0x050B; // 7E:050B
	byte Guard_frame              : "snes9x-x64.exe", 0x008D8C38, 0x0477; // 7E:0477
	ushort Guard_sequence_pointer : "snes9x-x64.exe", 0x008D8C38, 0x0480; // 7E:0480
	byte24 GdStartBlock           : "snes9x-x64.exe", 0x008D8C38, 0x1F114; // 7F:F114; guard tile positions for each room, changes to 0xFF if a guard falls offscreen
	byte Guard_type               : "snes9x-x64.exe", 0x008D8C38, 0x062B; // 7E:062B
	byte princess_image_time      : "snes9x-x64.exe", 0x008D8C38, 0x0E1C; // 7E:0E1C
}

startup
{
	//refreshRate = 30; // Prince of Persia runs at 12 fps, let's update 2.5x as often to be sure

	settings.Add("test_skip", false, "For testing: Quickly skip over each action if the game is past its level.");
	settings.Add("any_order", true, "Allow doing the actions in any order.");

	// This is the list of actions needed for 100% completion.
	// Keep this in sync with the splits!
	vars.checks = new int[][]
	{
		// columns: type, level, room, tile, done
		// types:
		// 'p' = pick up (potion, sword)
		// 'g' = guard: The tile is the guard type (for level 19) or -1 if we don't care.
		// Guard types are listed at: https://www.princed.org/wiki/SNES_format#Levels , however, you need to subtract 2 from the numbers listed there.
		// 'x' = level exit (ignores the room and the tile)
		// "done" starts off as 0 and is set to 1 when the action is done.
		// This column is there to support completing actions in any order.

		// Level 1
		new int[]{'p', 1,  8, 27, 0}, // Level 1: Heal potion at room 8, tile 27 (attic, under the guard)
		new int[]{'p', 1, 12,  4, 0}, // Level 1: Heal potion at room 12, tile 4 (cellar)
		new int[]{'p', 1, 20, 25, 0}, // Level 1: Sword at room 20, tile 25
		new int[]{'p', 1, 23,  1, 0}, // Level 1: Heal potion at room 23, tile 1 (attic left, above the skeleton)
		new int[]{'g', 1,  8, -1, 0}, // Level 1: Guard at room 8, tile 6 (attic)
		new int[]{'g', 1,  1, -1, 0}, // Level 1: Guard at room 1, tile 17 (before the exit)
		new int[]{'x', 1, -1, -1, 0}, // Level 1: Level exit

		// Level 2
		new int[]{'g', 2,  7, -1, 0}, // Level 2: Guard at room 7
		new int[]{'p', 2,  8, 14, 0}, // Level 2: Heal potion at room 2, tile 14 (room below the guard)
		new int[]{'g', 2, 11, -1, 0}, // Level 2: Guard at room 11 (still underground)
		new int[]{'p', 2,  5, 26, 0}, // Level 2: Heal potion at room 5, tile 26 (above the ceiling)
		new int[]{'g', 2,  2, -1, 0}, // Level 2: Guard at room 2 (at the skeleton)
		new int[]{'p', 2,  1, 27, 0}, // Level 2: Heal potion at room 1, tile 27 (between the gates)
		new int[]{'g', 2,  0, -1, 0}, // Level 2: Guard at room 0 (bluish, at the spikes)
		new int[]{'p', 2, 13, 28, 0}, // Level 2: Life potion at room 13, tile 28
		new int[]{'p', 2, 23, 25, 0}, // Level 2: Heal potion at room 23, tile 25 (ceiling)
		new int[]{'g', 2, 16, -1, 0}, // Level 2: Guard at room 16 (bluish)
		new int[]{'g', 2, 15, -1, 0}, // Level 2: Guard at room 15 (next to the exit)
		new int[]{'g', 2, 12, -1, 0}, // Level 2: Guard at room 12 (bluish) (under the button which opens the exit)
		new int[]{'x', 2, -1, -1, 0}, // Level 2: Level exit

		// Level 3
		new int[]{'g', 3, 21, -1, 0}, // Level 3: Guard at room 21
		new int[]{'g', 3,  9, -1, 0}, // Level 3: Brown Skeleton at room 9
		new int[]{'p', 3,  9, 27, 0}, // Level 3: Life potion at room 9, tile 27 (at the brown skeleton)
		new int[]{'p', 3,  8,  2, 0}, // Level 3: Life potion at room 8, tile 2 (behind chompers)
		new int[]{'p', 3, 10,  8, 0}, // Level 3: Heal potion at room 10, tile 8 (after the first gate)
		new int[]{'p', 3, 18, 11, 0}, // Level 3: Hurt potion at room 18, tile 11 (on the way to the exit opener)
		new int[]{'p', 3, 18, 21, 0}, // Level 3: Heal potion at room 18, tile 21
		new int[]{'p', 3, 12, 16, 0}, // Level 3: Heal potion at room 12, tile 16 (underground)
		new int[]{'g', 3, 13, -1, 0}, // Level 3: Skeleton at room 13, crushed in room 20, or dropped offscreen from room 11
		new int[]{'x', 3, -1, -1, 0}, // Level 3: Level exit

		// Level 4
		new int[]{'g', 4,  8, -1, 0}, // Level 4: Guard at room 8
		new int[]{'p', 4, 13, 12, 0}, // Level 4: Heal potion at room 13, tile 12 (behind the guard)
		new int[]{'p', 4, 12, 22, 0}, // Level 4: Heal potion at room 12, tile 22 (ceiling)
		new int[]{'g', 4,  2, -1, 0}, // Level 4: Guard at room 2
		new int[]{'g', 4,  3, -1, 0}, // Level 4: Guard at room 3
		new int[]{'p', 4, 17, 15, 0}, // Level 4: Heal potion at room 17, tile 15 (ceiling)
		new int[]{'p', 4, 17, 17, 0}, // Level 4: Heal potion at room 17, tile 17 (ceiling)
		new int[]{'g', 4, 16, -1, 0}, // Level 4: Guard at room 16
		new int[]{'p', 4, 22, 16, 0}, // Level 4: Life potion at room 22, tile 16
		new int[]{'x', 4, -1, -1, 0}, // Level 4: Level exit

		// Level 5
		new int[]{'p', 5,  8, 21, 0}, // Level 5: Heal potion at room 8, tile 21
		new int[]{'g', 5, 11, -1, 0}, // Level 5: Guard at room 11
		new int[]{'p', 5, 13, 10, 0}, // Level 5: Heal potion at room 13, tile 10
		new int[]{'g', 5, 13, -1, 0}, // Level 5: Guard at room 13
		new int[]{'g', 5,  2, -1, 0}, // Level 5: Guard at room 2 (outdoor)
		new int[]{'p', 5,  3,  2, 0}, // Level 5: Heal potion at room 3, tile 2
		new int[]{'p', 5, 16, 12, 0}, // Level 5: Heal potion at room 16, tile 12
		new int[]{'p', 5, 21, 13, 0}, // Level 5: Life potion at room 21, tile 13 (hidden)
		new int[]{'x', 5, -1, -1, 0}, // Level 5: Level exit

		// Level 6
		new int[]{'p', 6, 10,  4, 0}, // Level 6: Heal potion at room 10, tile 4 (behind chompers)
		new int[]{'g', 6, 11, -1, 0}, // Level 6: Guard at room 11
		new int[]{'g', 6,  0, -1, 0}, // Level 6: Guard at room 0
		new int[]{'g', 6,  1, -1, 0}, // Level 6: Guard at room 1
		new int[]{'p', 6,  6,  8, 0}, // Level 6: Heal potion at room 6, tile 8
		new int[]{'p', 6, 14, 15, 0}, // Level 6: Life potion at room 14, tile 15
		new int[]{'p', 6,  5, 18, 0}, // Level 6: Hurt potion at room 5, tile 18
		new int[]{'g', 6, 21, -1, 0}, // Level 6: Amazon at room 21
		new int[]{'x', 6, -1, -1, 0}, // Level 6: Level exit

		// Level 7
		new int[]{'p', 7,  0,  7, 0}, // Level 7: Heal potion at room 0, tile 7
		new int[]{'g', 7,  8, -1, 0}, // Level 7: Guard at room 8
		new int[]{'p', 7,  6, 24, 0}, // Level 7: Heal potion at room 6, tile 24
		new int[]{'p', 7,  6, 25, 0}, // Level 7: Heal potion at room 6, tile 25
		new int[]{'g', 7, 11, -1, 0}, // Level 7: Guard at room 11
		new int[]{'p', 7, 11, 25, 0}, // Level 7: Heal potion at room 11, tile 25
		new int[]{'p', 7, 12,  6, 0}, // Level 7: Heal potion at room 12, tile 6
		new int[]{'p', 7, 10, 17, 0}, // Level 7: Heal potion at room 10, tile 17
		new int[]{'p', 7, 16,  6, 0}, // Level 7: Life potion at room 16, tile 6 (stolen)
		new int[]{'p', 7, 13,  0, 0}, // Level 7: Heal potion at room 13, tile 0
		new int[]{'p', 7, 13,  1, 0}, // Level 7: Heal potion at room 13, tile 1
		new int[]{'p', 7, 13,  2, 0}, // Level 7: Heal potion at room 13, tile 2
		new int[]{'p', 7, 13,  3, 0}, // Level 7: Heal potion at room 13, tile 3
		new int[]{'p', 7, 20, 13, 0}, // Level 7: Life potion at room 20, tile 13 (hidden)
		new int[]{'g', 7, 19, -1, 0}, // Level 7: Guard at room 19
		new int[]{'x', 7, -1, -1, 0}, // Level 7: Level exit

		// Level 8
		new int[]{'p', 8, 23,  5, 0}, // Level 8: Life potion at room 23, tile 5 (hidden)
		new int[]{'p', 8,  6,  7, 0}, // Level 8: Heal potion at room 6, tile 7
		new int[]{'g', 8,  2, -1, 0}, // Level 8: Guard at room 2
		new int[]{'g', 8, 11, -1, 0}, // Level 8: Guard at room 11
		new int[]{'p', 8, 11, 25, 0}, // Level 8: Heal potion at room 11, tile 25
		new int[]{'g', 8, 17, -1, 0}, // Level 8: Guard at room 17
		new int[]{'g', 8, 21, -1, 0}, // Level 8: Guard at room 21
		new int[]{'x', 8, -1, -1, 0}, // Level 8: Level exit

		// Level 9
		new int[]{'g', 9,  1, -1, 0}, // Level 9: Guard at room 1
		new int[]{'g', 9,  5, -1, 0}, // Level 9: Guard at room 5
		new int[]{'p', 9, 19, 25, 0}, // Level 9: Heal potion at room 19, tile 25
		new int[]{'p', 9, 15, 14, 0}, // Level 9: Life potion at room 15, tile 14 (hidden)
		new int[]{'g', 9, 16, -1, 0}, // Level 9: Politician (fat guard) at room 16
		new int[]{'x', 9, -1, -1, 0}, // Level 9: Level exit

		// Level 10
		new int[]{'g', 10,  0, -1, 0}, // Level 10: Skeleton at room 0
		new int[]{'g', 10,  2, -1, 0}, // Level 10: Skeleton at room 2
		new int[]{'g', 10,  4, -1, 0}, // Level 10: Brown skeleton at room 4
		new int[]{'g', 10,  5, -1, 0}, // Level 10: Skeleton at room 5
		new int[]{'p', 10, 19, 14, 0}, // Level 10: Life potion at room 19, tile 14 (hidden)
		new int[]{'g', 10,  7, -1, 0}, // Level 10: Skeleton at room 7
		new int[]{'g', 10, 10, -1, 0}, // Level 10: Skeleton at room 10
		new int[]{'g', 10, 14, -1, 0}, // Level 10: Skeleton at room 14
		new int[]{'g', 10, 16, -1, 0}, // Level 10: Brown skeleton at room 16
		new int[]{'x', 10, -1, -1, 0}, // Level 10: Level exit

		// Level 11
		new int[]{'p', 11,  5,  2, 0}, // Level 11: Heal potion at room 5, tile 2 (top left)
		new int[]{'g', 11,  6, -1, 0}, // Level 11: Guard at room 6
		new int[]{'g', 11,  9, -1, 0}, // Level 11: Guard at room 9 (above a potion)
		new int[]{'p', 11,  9, 27, 0}, // Level 11: Heal potion at room 9, tile 27 (under a guard)
		new int[]{'g', 11, 13, -1, 0}, // Level 11: Guard at room 13
		new int[]{'p', 11,  7, 12, 0}, // Level 11: Heal potion at room 7, tile 12
		new int[]{'g', 11, 11, -1, 0}, // Level 11: Guard at room 11 (at the chomper)
		new int[]{'p', 11, 16, 12, 0}, // Level 11: Heal potion at room 16, tile 12 (under the spikes)
		new int[]{'p', 11, 19, 17, 0}, // Level 11: Life potion at room 19, tile 17
		new int[]{'x', 11, -1, -1, 0}, // Level 11: Level exit

		// Level 12
		new int[]{'g', 12,  2, -1, 0}, // Level 12: Guard at room 2
		new int[]{'g', 12,  7, -1, 0}, // Level 12: Guard at room 7
		new int[]{'g', 12, 21, -1, 0}, // Level 12: Guard at room 21
		new int[]{'p', 12, 21, 28, 0}, // Level 12: Heal potion at room 21, tile 28
		new int[]{'g', 12, 17, -1, 0}, // Level 12: Guard at room 17
		new int[]{'p', 12, 19, 20, 0}, // Level 12: Heal potion at room 19, tile 20
		new int[]{'p', 12, 19, 21, 0}, // Level 12: Heal potion at room 19, tile 21
		new int[]{'g', 12, 14, -1, 0}, // Level 12: Guard at room 14
		new int[]{'p', 12,  1, 20, 0}, // Level 12: Slow fall potion at room 1, tile 20
		new int[]{'p', 12,  5, 10, 0}, // Level 12: Life potion at room 5, tile 10
		new int[]{'g', 12,  4, -1, 0}, // Level 12: Golden skeleton at room 4
		new int[]{'x', 12, -1, -1, 0}, // Level 12: Level exit

		// Level 13
		new int[]{'p', 13,  2, 11, 0}, // Level 13: Hurt potion at room 2, tile 11
		new int[]{'g', 13,  3, -1, 0}, // Level 13: Guard at room 3
		new int[]{'p', 13,  3, 25, 0}, // Level 13: Heal potion at room 3, tile 25
		new int[]{'g', 13,  4, -1, 0}, // Level 13: Guard at room 4
		new int[]{'g', 13, 13, -1, 0}, // Level 13: Guard at room 13
		new int[]{'g', 13,  6, -1, 0}, // Level 13: Guard at room 6
		new int[]{'p', 13,  0, 10, 0}, // Level 13: Life potion at room 0, tile 10
		new int[]{'g', 13, 18, -1, 0}, // Level 13: Guard at room 18
		new int[]{'x', 13, -1, -1, 0}, // Level 13: Level exit

		// Level 14
		new int[]{'g', 14,  1, -1, 0}, // Level 14: Guard at room 1
		new int[]{'p', 14, 21, 17, 0}, // Level 14: Heal potion at room 21, tile 17
		new int[]{'g', 14,  4, -1, 0}, // Level 14: Guard at room 4
		new int[]{'p', 14,  5,  2, 0}, // Level 14: Heal potion at room 5, tile 2
		new int[]{'g', 14, 10, -1, 0}, // Level 14: Guard at room 10
		new int[]{'p', 14, 23, 27, 0}, // Level 14: Life potion at room 23, tile 27 (hidden)
		new int[]{'g', 14, 12, -1, 0}, // Level 14: Guard at room 12
		new int[]{'p', 14, 14, 27, 0}, // Level 14: Hurt potion at room 14, tile 27
		new int[]{'p', 14, 15, 17, 0}, // Level 14: Flip potion at room 15, tile 17
		new int[]{'p', 14, 20, 18, 0}, // Level 14: Life potion at room 20, tile 18
		new int[]{'p', 14, 18, 11, 0}, // Level 14: Flip potion at room 18, tile 11
		new int[]{'x', 14, -1, -1, 0}, // Level 14: Level exit

		// Level 15
		new int[]{'g', 15,  0, -1, 0}, // Level 15: Guard at room 0 (purple)
		new int[]{'g', 15,  3, -1, 0}, // Level 15: Guard at room 3 (blue)
		new int[]{'p', 15, 22, 16, 0}, // Level 15: Heal potion at room 22, tile 16
		new int[]{'g', 15,  5, -1, 0}, // Level 15: Guard at room 5 (red)
		new int[]{'g', 15, 10, -1, 0}, // Level 15: Guard at room 10 (purple)
		new int[]{'p', 15, 22, 11, 0}, // Level 15: Life potion at room 22, tile 11 (hidden)
		new int[]{'p', 15,  8, 27, 0}, // Level 15: Heal potion at room 8, tile 27
		new int[]{'g', 15, 17, -1, 0}, // Level 15: Shadow at room 17
		new int[]{'x', 15, -1, -1, 0}, // Level 15: Level exit

		// Level 16
		new int[]{'g', 16,  3, -1, 0}, // Level 16: Guard at room 3 (blue)
		new int[]{'g', 16,  4, -1, 0}, // Level 16: Guard at room 4 (blue)
		new int[]{'p', 16,  8,  2, 0}, // Level 16: Life potion at room 8, tile 2
		new int[]{'g', 16, 10, -1, 0}, // Level 16: Guard at room 10 (purple)
		new int[]{'p', 16,  5, 25, 0}, // Level 16: Warp potion at room 5, tile 25
		new int[]{'g', 16, 14, -1, 0}, // Level 16: Guard at room 14 (red)
		new int[]{'g', 16, 20, -1, 0}, // Level 16: Red knight at room 20
		new int[]{'x', 16, -1, -1, 0}, // Level 16: Level exit

		// Level 17
		new int[]{'p', 17, 21, 25, 0}, // Level 17: Life potion at room 21, tile 25 (hidden)
		new int[]{'g', 17,  2, -1, 0}, // Level 17: Six-armed guy at room 2
		new int[]{'x', 17, -1, -1, 0}, // Level 17: Level exit

		// Level 18
		new int[]{'p', 18, 18, 21, 0}, // Level 18: Heal potion at room 18, tile 21
		new int[]{'g', 18, 16, -1, 0}, // Level 18: Guard at room 16 (purple)
		new int[]{'g', 18, 23, -1, 0}, // Level 18: Brown skeleton at room 23
		new int[]{'p', 18, 23, 25, 0}, // Level 18: Life potion at room 23, tile 25
		new int[]{'p', 18, 23, 23, 0}, // Level 18: Kill potion at room 23, tile 23
		new int[]{'p', 18,  8,  8, 0}, // Level 18: Heal potion at room 8, tile 8
		new int[]{'g', 18, 22, -1, 0}, // Level 18: Guard at room 22 (purple)
		new int[]{'g', 18, 21, -1, 0}, // Level 18: Guard at room 21 (red)
		new int[]{'p', 18, 21, 18, 0}, // Level 18: Heal potion at room 21, tile 18
		new int[]{'g', 18, 11, -1, 0}, // Level 18: Blue knight at room 11
		new int[]{'g', 18, 14, -1, 0}, // Level 18: Guard at room 14 (purple)
		new int[]{'x', 18, -1, -1, 0}, // Level 18: Level exit

		// Level 19
		new int[]{'p', 19, 11, 18, 0}, // Level 19: Life potion at room 11, tile 18
		new int[]{'g', 19,  7, 12, 0}, // Level 19: Guard at room 7 (blue face)
		new int[]{'g', 19,  7,  3, 0}, // Level 19: Golden skeleton at room 7
		new int[]{'g', 19,  7,  4, 0}, // Level 19: Amazon at room 7
		new int[]{'g', 19,  7,  5, 0}, // Level 19: Politician (fat guard) at room 7
		new int[]{'g', 19,  7, 14, 0}, // Level 19: Blue knight at room 7
		new int[]{'x', 19, -1, -1, 0}, // Level 19: Level exit

		// Level 20
		new int[]{'i', 20, -1, -1, 0}, // Level 20: See the princess on the sky
		new int[]{'x', 20, -1, -1, 0}, // Level 20: Level exit (Jaffar warps you to the final battle)
		new int[]{'g', 21,  0, -1, 0}, // Level 21: Defeat Jaffar
		new int[]{'x', 21, -1, -1, 0}, // Level 21: Level exit (warped back to level 20)
		new int[]{'x', 20, -1, -1, 0}, // Level 20: Level exit (run into the princess's room)
		// SOLVED: "Run into the princess's room" is triggered immediately after "back to level 20".
	};

	vars.check_index = 0;
	vars.playing = false;

	vars.reset_guards = (Action)delegate()
	{
		int[] array = new int[24];
		for (int i = 0; i < 24; i++) array[i] = i;
		vars.guard_original_rooms = array;
	};

	vars.reset_actions = (Action)delegate()
	{
		// Mark all actions as not done.
		for (int check_index = 0; check_index < vars.checks.Length; check_index++)
		{
			vars.checks[check_index][4] = 0;
		}
	};

	vars.calculate_time = (Func<int,int,TimeSpan>)delegate(int FrameCount, int FrameCountHalf)
	{
		// 1 minute = 0x01A9 frames, 1 second = 7 frames.
		// So 1 minute is slightly more than 60 seconds, and they don't even match real time.

		int overflow_bits = (FrameCountHalf>>1)<<16;
		int half_frame = FrameCountHalf & 1;

		return TimeSpan.FromMinutes( ((double)(overflow_bits + FrameCount) + 0.5 * (double)half_frame) / (double)0x01A9);
	};

	vars.reset_guards();
	// Reset if this autosplitter is (changed and) reloaded, because then we lose track of the current split anyway.
	vars.reset = true;
}

reset
{
	// KidMaxHP == 0 when the game is reset.
	// CurrentLevel == 0 on level 1 *and* in the main menu!
	// CurrentLevel >= 21 are the demo and training levels.
	// CurrentLevel == 128 during cutscenes.
	// CurrentLevel == 252 after the demo level was interrupted and before the main menu fully appears.
	// CurrentLevel == 254 after GAME END was selected and before the main menu fully appears.
	// CurrentLevel == 255 when restarting the level.
	bool reset = (
		current.KidMaxHP == 0
		|| (current.CurrentLevel >= 21 && current.CurrentLevel < 128)
		|| current.CurrentLevel == 252
		|| current.CurrentLevel == 254
		|| vars.reset
	);
	if (reset)
	{
		print("reset");
		print("CurrentLevel = " + current.CurrentLevel);
		vars.check_index = 0;
		vars.playing = false;
		vars.reset_guards();
		vars.reset_actions();
		vars.reset = false;
	}
	return reset;
}

start
{
	// This should become true only when the player selects START GAME.
	// We check StartRoom to prevent false positives, as otherwise the condition is true after the player quits the demo/training level.
	// PROBLEM: This doesn't work if I fast-forward through START GAME, because then FrameCount is 0 only for a too short time.
	bool start = (
		current.KidMaxHP == 3
		&& current.FrameCount == 0
		&& current.CurrentLevel == 0
		&& current.StartRoom == 22
		//&& current.StartRoom == current.Kid_room
	);
	if (start)
	{
		print("start");
		vars.check_index = 0;
		vars.playing = true;
		vars.reset_guards();
		vars.reset_actions();
	}
	return start;
}

update
{
	if (!vars.playing) return;

	//print("settings: any_order = " + settings["any_order"]);
	//print("settings: test_skip = " + settings["test_skip"]);
	//print("vars.check_index = " + vars.check_index);

	// CurrentLevel == 255 when restarting the level.
	if (current.CurrentLevel == 255) vars.reset_guards();

	// CurrentLevel == 128 during cutscenes.
	if (current.CurrentLevel >= 128) return;

	//print("It's alive!");
	//print("CurrentLevel = "+current.CurrentLevel);
	//print("NextLevel = "+current.NextLevel);
	//print("KidMaxHP = "+current.KidMaxHP);

	// If a guard moved to another room:
	if (
		( // regular guards
			current.Guard_direction != 0x7F
			&& old.Guard_direction != 0x7F
			&& current.Guard_room != old.Guard_room
			&& current.Guard_direction == old.Guard_direction
			&& current.Guard_hp == old.Guard_hp
			// TODO: Check more attributes of the guard to make sure they are the same guard?
			// TODO: Check if the prince is in fight mode? Or, can guards move between rooms even if he isn't?
		)
		||
		( // When the level 3 skeleton falls offscreen, it disappears *but* may reappear in another room!
			current.Guard_direction == 0x7F
			&& old.Guard_direction != 0x7F
			&& current.Guard_room != old.Guard_room
			&& (current.GdStartBlock[current.Guard_room] != 0xFF && old.GdStartBlock[current.Guard_room] == 0xFF)
		)
	)
	{
		// If there is a guard in room R, then guard_original_rooms[R] tells the room where that guard started the level.
		int guard_original_room = vars.guard_original_rooms[old.Guard_room];
		print("A guard moved from room "+old.Guard_room+" to "+current.Guard_room+". (original room: "+guard_original_room+")");
		vars.guard_original_rooms[current.Guard_room] = guard_original_room;
	}

	// Stop if the player already completed all actions.
	if (vars.check_index >= vars.checks.Length) return;

	// DONE: Allow doing the actions in any order within a level.
	int first_index, last_index;
	if (settings["any_order"])
	{
		// Look at all actions.
		// Ideally we would only look at the action of the current level...
		first_index = vars.check_index;
		last_index = vars.checks.Length - 1;
	}
	else
	{
		// Look at only the current action.
		first_index = last_index = vars.check_index;
	}

	for (int check_index = first_index; check_index <= last_index; check_index++)
	{
		// If a check matches, mark it as done (in a new check[4]).
		// If vars.checks[vars.check_index] is done then mark a split.

		//int[] check = vars.checks[vars.check_index];
		int[] check = vars.checks[check_index];
		char type = (char)check[0];
		int level = check[1];
		int room = check[2];
		int tile = check[3];
		//print("check_index = "+vars.check_index);
		//print("type = "+type);
		//print("level = "+level);
		//print("room = "+room);
		//print("tile = "+tile);

		if (settings["test_skip"])
		{
			// For testing: Quickly skip over each action if the game is past its level.
			// DONE: Added a setting for this.
			// Without "level < 20", "Run into the princess's room" triggers too early.
			if (current.CurrentLevel+1 > level && level < 20)
			{
				check[4] = 1;
				vars.reset_guards();
				continue;
			}
		}

		// Attempt to fix "Run into the princess's room" triggering too early.
		if (current.CurrentLevel+1 < level) break;

		// Type IDs of objects which can be picked up:
		// (Source: https://www.princed.org/wiki/SNES_format#Levels)
		// 0x0A = sword
		// 0x12 = heal potion
		// 0x13 = life potion
		// 0x14 = hurt potion
		// 0x15 = upside down potion
		// 0x16 = slow fall potion
		// 0x1A = potion of warp
		// 0x1C = kill potion

		if (current.CurrentLevel+1 == level)
		{

			switch (type)
			{
				case 'p': // pick up
					if (current.CurrentLevel+1 == level)
					{
						int offset = room*30+tile;
						// TODO: Compare the old value against the specific object type?
						// SOLVED: If we check the old value in any way, then we won't detect actions if the player does them out of order.
						// PROBLEM: If we *don't* check the old value, we can get false positives before the level is fully loaded.
						if (current.LevelObjects[offset] == 0x00 && old.LevelObjects[offset] != 0x00)
						{
							print("The player picked up the item at: level "+level+", room "+room+", tile "+tile);
							check[4] = 1;
						}
					}
				break;

				case 'g': // guard
					if (
						(
							// the guard died onscreen
							current.Guard_direction != 0x7F // a guard is present
							&& (
								(current.Guard_alive == 0 && old.Guard_alive != 0) // died
								// We need to check these separately because skeletons are not marked as dead in Guard_alive.
								|| (current.Guard_frame == 184 && old.Guard_frame != 184) // level 3 skeleton crushed
								|| (old.Guard_frame == 177 && current.Guard_frame == 186) // brown skeleton collapsed
							)
						)
						||
						(
							// the guard fell out of the room
							(current.Guard_direction == 0x7F && old.Guard_direction != 0x7F) // the guard disappeared
							//&& (current.Guard_room == old.Guard_room) // PROBLEM: This is false for the level 3 skeleton.
							&& (current.Kid_room == old.Kid_room) // not because the kid moved to another room
							&& (current.GdStartBlock[old.Guard_room] == 0xFF /*&& old.GdStartBlock[old.Guard_room] != 0xFF*/) // guard cleared from the old room
							&& !(current.GdStartBlock[current.Guard_room] != 0xFF && old.GdStartBlock[current.Guard_room] == 0xFF) // guard did not reappear in another room (as the level 3 skeleton does)
							// Without the current.GdStartBlock[old.Guard_room] condition, this sometimes becomes true when the player merely leaves the room of a living guard!
							// Because the guard is removed before Kid_room changes.
							// Don't check old.GdStartBlock[old.Guard_room], because it *is* 0xFF if the guard did not start in the room where he fell offscreen.
						)
					)
					{
						print("current.Guard_direction = " + current.Guard_direction);
						print("old.Guard_direction = " + old.Guard_direction);
						print("current.Guard_room = " + current.Guard_room);
						print("old.Guard_room = " + old.Guard_room);
						print("current.Kid_room = " + current.Kid_room);
						print("old.Kid_room = " + old.Kid_room);
						print("current.GdStartBlock[current.Guard_room] = " + current.GdStartBlock[current.Guard_room]);
						print("old.GdStartBlock[current.Guard_room] = " + old.GdStartBlock[current.Guard_room]);
						int guard_original_room = vars.guard_original_rooms[old.Guard_room];
						print("A guard died in room " + old.Guard_room + " of level " + (current.CurrentLevel+1)
							+ ". (original room: "+guard_original_room+")");
						if (current.CurrentLevel+1 == level && guard_original_room == room
							// For guards, "tile" stores the guard type or -1.
							&& (tile == -1 || tile == old.Guard_type)
						)
						{
							check[4] = 1;
						}
					}
				break;

				case 'x': // level exit
					if (current.CurrentLevel+1 == level
						&& current.NextLevel != old.NextLevel
						&& old.NextLevel == old.CurrentLevel
						&& current.NextLevel != current.CurrentLevel)
					{
						check[4] = 1;
						vars.reset_guards();
						print("Level exit from level "+(current.CurrentLevel+1)+" to level " + (current.NextLevel+1) + ". (Event #"+(check_index)+")");
						// Attempt to fix "Run into the princess's room" triggering too early.
						return;
					}
				break;

				case 'i': // Princess image
					if (current.princess_image_time >= 100)
					{
						check[4] = 1;
					}
				break;
			}

		}

	}

}

split
{
	int[] check = vars.checks[vars.check_index];
	bool split = (check[4] == 1);
	if (split) vars.check_index++;
	return split;
}

gameTime
{
	if (current.CurrentLevel+1 >= 20)
	{
		// The main timer stops on level 20, but the per-level counter keeps counting.
		// PROBLEM: FrameCountThisLevel resets when you are warped back to level 20.
		return vars.calculate_time(current.FrameCountLevelStart, current.FrameCountHalfLevelStart)
		     + vars.calculate_time(current.FrameCountThisLevel , current.FrameCountHalfThisLevel );
	}
	else
	{
		//return TimeSpan.FromMinutes( ((double)current.FrameCount + 0.5 * (double)current.FrameCountHalf) / (double)0x01A9);
		return vars.calculate_time(current.FrameCount, current.FrameCountHalf);
	}
}

isLoading
{
	// Force Game Time to show the actual Game Time...
	return true;
}
